/*********************************************************/
/* file: mapdata.c                                       */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:   1 april 93                                 */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  if you add place your name here or below             */
/*                                                       */
/* description:                                          */
/*  mapdata.c contains subroutines to handle the dataflow*/
/* on the maps, and of the maps to and from the tmp dir  */
/*                                                       */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <obdata.h> 

/**********************************************/
/* maintain_maps                              */
/*    this will run through the maps searching*/
/* for maps that have not been referenced and */
/* marking them idle, maps that have been idle*/
/* too long are paged out to the tmp device   */
/**********************************************/
void
maintain_maps()
{
  map            *current,
                 *next;
  
  for (current = maps; current != NULL; current = next)
    {
      next = current->next;
      /* if map had no refs increase idle_count */
      if (current->refs == 0)
	current->idle++;
      else
	current->idle = 0;
      /* if idle too long page map out */
      if (current->idle > MAX_MAP_IDLE)
	{
	  /* save to the tmp dir */
	  /*fprintf (stderr,"paging out\n");*/
	  save_map_data(current, TO_TMP_DEV);
	  /* remove it from memory */
	  deallocate_map(current);
      } else
	/* reset reference counter */
	current->refs = 0;
    }
}
 
/*****************************************/
/* deallocate map                        */
/*   used to 'free' the objects in a map */
/* will free the map data structure as   */
/* well as the tile structures and       */
/* objects associated with that map      */
/*****************************************/
void
deallocate_map(map * amap)
{
  map           **curr;
  int             x,
                  z;
  int             mapsize;

  object         *tmp_ob,
                 *next;
  
  mapsize = amap->boundx*amap->boundy;
  for (x = 0; x < mapsize; x++)
    for (z = 0; z < 3; z++)
      /* remember that freeing an object removes it from the map ! */
      /* hence some of the pointers are no longer valid            */
      /* I am considering hacking the removed variable to make     */
      /* this subroutine run faster                                */
      for (tmp_ob = amap->tiles[x].vol[z].top; tmp_ob != NULL;
	   tmp_ob = next)
	{
	  next = tmp_ob->below;
	  
	  /* the hack....   */
	  SET_REMOVED(tmp_ob);
	  
	  free_object(tmp_ob);
	}
  free (amap->tiles);
  /* actually i could compare the pointers here instead */
  for (curr = &maps; ((*curr) != NULL) && (strcmp((*curr)->name, amap->name));
       curr = &((*curr)->next));
  if (*curr == NULL)
    fprintf(stderr, "error in map deallocation");
  *curr = (*curr)->next;
  free(amap);
}

map            *
allocate_map(int mapwidth, int maplength, char *name)
{
  map            *newmap;
  int             mapsize;
  
  mapsize = mapwidth * maplength;
  newmap = (map *) malloc(sizeof(map));
  memset(newmap, 0, sizeof(map));
  newmap->boundx = mapwidth;
  newmap->boundy = maplength;
  newmap->refs = newmap->idle = 0;
  newmap->tiles = (square *) calloc(mapsize, sizeof(square));
  strcpy(newmap->name, name);
  newmap->next = maps;
  maps = newmap;
  return (newmap);
}

/*******************************************/
/* insert_ob_in_map                        */
/*    insert_ob_in_map will change an      */
/*  objects data structure to reflect the  */
/*  presence of the object in the given map*/
/*  at the x y z coordinates supplied to   */
/*  the function.                          */
/*******************************************/
void
insert_ob_in_map(map * amap, object * obj, int x, int y, int z)
{
  object *curr;
  obj = HEAD(obj);
  /* check for bad insert coords */
  if (!IN_MAP(amap, x, y))
    {
      /*fprintf(stderr, "tried to insert object outside of map\n");*/
      return;
    }
  
  for (curr = obj;  curr != NULL && z < 3 ; curr = curr->link, z++)
    {
      /* empty squares require special linking */
      if (EMPTY(amap, x, y, z))
	{
	  BOT_MAP_OBJ(amap, x, y, z) = curr;
	  TOP_MAP_OBJ(amap, x, y, z) = curr;
	  curr->below = NULL;
	  curr->above = NULL;
      } else
	/* otherwise perform normal linking */
	{
	  TOP_MAP_OBJ(amap, x, y, z)->above = curr;
	  curr->below = TOP_MAP_OBJ(amap, x, y, z);
	  TOP_MAP_OBJ(amap, x, y, z) = curr;
	  curr->above = NULL;
	}
      
      /* change objects data structure to reflect changes */
      curr->x = x;
      curr->y = y;
      curr->z = z;
      curr->mapin = amap;
      
      /* if it is a floor, then set the floor pointer for */
      /* that maps x,y to this object                     */
      if (curr->type == floor && (z == 0))
	MAP_FLOOR_OBJ(amap, x, y) = curr;
    }
  
  CLEAR_REMOVED(obj); 
}

/*******************************************/
/* insert_ob_under_map                     */
/*    insert_ob_under_map will change an   */
/*  objects data structure to reflect the  */
/*  presence of the object in the given map*/
/*  at the x y z coordinates supplied to   */
/*  the function. note that this function  */
/*  differs from insert_ob_in_map in that  */
/*  it inserts the object underneath the   */
/*  topmost object at the given maps x y z */
/*******************************************/
void
insert_ob_under_map(map * amap, object * obj, int x, int y, int z)
{
	object *curr;
	
	obj = HEAD(obj); 
  /* check for bad insert coords */
  if (!IN_MAP(amap, x, y))
  {
    /*fprintf(stderr, "tried to insert object outside of map\n");*/
        return;
  }
  /* empty squares require special linking */
  
  for (curr = obj;  curr != NULL && z < 3; curr = curr->link, z++)
  {
  if (EMPTY(amap, x, y, z))
  {
    BOT_MAP_OBJ(amap, x, y, z) = curr;
    TOP_MAP_OBJ(amap, x, y, z) = curr;
    curr->below = NULL;
    curr->above = NULL;
  } else 
    /* otherwise perform normal linking by inserting */
    /* the object below the topmost object in map    */
    /* if the area is blocked                        */
  if (MAP_MOVE_BLOCKED(amap,x,y,z))
  {
    curr->below = TOP_MAP_OBJ(amap, x, y, z)->below;
    if (TOP_MAP_OBJ(amap, x, y, z)->below != NULL)
      TOP_MAP_OBJ(amap, x, y, z)->below->above = curr;
    TOP_MAP_OBJ(amap, x, y, z)->below = curr;
    curr->above = TOP_MAP_OBJ(amap, x, y, z);
    if (curr->type == 223 && (z == 0) && !IS_FLOOR(curr->above))
    MAP_FLOOR_OBJ(amap, x, y) = curr;

   }
 	/* otherwise perform normal linking */
	else
  {
    TOP_MAP_OBJ(amap, x, y, z)->above = curr;
    curr->below = TOP_MAP_OBJ(amap, x, y, z);
    TOP_MAP_OBJ(amap, x, y, z) = curr;
    curr->above = NULL;
        if (curr->type == floor && (z == 0))
    MAP_FLOOR_OBJ(amap, x, y) = curr;


  }

  /* change objects data structure to reflect changes */
  curr->x = x;
  curr->y = y;
  curr->z = z;
  curr->mapin = amap;
  /* if it is a floor, then set the floor pointer for */
  /* that maps x,y to this object                     */
    }
  CLEAR_REMOVED(obj);
}



