/*********************************************************/
/* file: object.c                                        */
/*                                                       */
/* Copyright (C) 1994 Karl A Holland                     */
/*                                                       */
/* author: Karl A Holland                                */
/*                                                       */
/* created:   1 april 93                                 */
/*                                                       */
/* modified:  7 february 94                              */
/*                                                       */
/* contributing authors:                                 */
/*  if you add place your name here or below             */
/*                                                       */
/* description:                                          */
/*  object.c contains subroutines that are on a higher   */
/* level than obdata.c, but lower than those found in    */
/* subs.c. it implements many of the subroutines that are*/
/* related to game mechanics via calls to other routines */
/* within this file and to the lower level subroutines   */
/*                                                       */
/* LICENSING:                                            */
/*   a full license  is provided in the file LICENSE     */
/* which should have been included with this program when*/
/* it was copied. It may be found in the top directory   */
/* of any NORMAL distribution, named LICENSE             */
/*********************************************************/
#include <object.h>


/*******************************************/
/* send_to_map                             */
/*   essentially implements the neccessary */
/* steps to send an object from one map to */
/* another.                                */
/*******************************************/
int
send_to_map(object * obj, char *name, int x, int y, int z)
{
  map           **curr,
                 *rmap;

  obj = HEAD(obj);
  /* check to see if map loaded already */
  for (curr = &maps; ((*curr) != NULL) && (strcmp((*curr)->name, name));
       curr = &((*curr)->next));

  if (*curr != NULL)
  {
    remove_ob(obj);
    insert_ob_in_map(*curr, obj, x, y, z);
  } else
  {

    rmap = (read_map_data(name));
    if (rmap != NULL)
    {
      remove_ob(obj);
      insert_ob_in_map(rmap, obj, x, y, z);
    } else
      return (0);
  }
  return (1);
}

/*****************************************/
/* distance                              */
/* calculates the square of the distance */
/* between two objects. Please avoid     */
/* using calls to sqrt and such for speed*/
/*****************************************/
int
distance(object * ob1, object * ob2)
{
  int             x1 = ob1->x,
                  x2 = ob2->x,
                  y1 = ob1->y,
                  y2 = ob2->y;

  /* euclidean distance = (x1-x2)^2 + (y1-y2)^2 */
  return (((x1 - x2) * (x1 - x2)) + ((y1 - y2) * (y1 - y2)));
}

/*************************************/
/* find_nearest_player               */
/*    finds the nearest player to a  */
/* given object via the distance     */
/* function                          */
/*************************************/
object         *
find_nearest_player(object * ob)
{
  object         *current,
                 *nearest;

  int             lowest;

  if (players == NULL)
    return NULL;

  for (current = players, nearest = NULL, lowest = 1 << 30; IS_PLAYER(current);
       current = current->next)
    if ((distance(current, ob) < lowest) && (current->mapin == ob->mapin))
    {
      lowest = distance(current, ob);
      nearest = current;
    }
  return (nearest);
}

/******************************************/
/* find direction                         */
/*    calculates the best direction for an*/
/* object to move in order to get to the  */
/* other object (from ob1 to ob2)         */
/******************************************/
int
find_direction(object * ob1, object * ob2)
{
  int             tdir = 4;
  if (ob2 == NULL || ob1 == NULL)
    return -1;
  if (ob1->x < ob2->x)
    tdir += 1;
  else if (ob1->x > ob2->x)
    tdir -= 1;
  if (ob1->y < ob2->y)
    tdir -= 3;
  else if (ob1->y > ob2->y)
    tdir += 3;
  return dirtrans[tdir];
}

/***********************************/
/* move_object                     */
/*  will attempt to move a given   */
/* object in the given direction   */
/* if movement and phaseability    */
/* allow it to. returns non-zero   */
/* on success                      */
/***********************************/
int
move_object(object * obj, int dir)
{
  int             newx,
                  newy;

  object         *curr;

  obj = HEAD(obj);

  if (dir < 0 || OB_IS_REMOVED(obj))
    return 0;

  newx = obj->x + xdeltas[dir];
  newy = obj->y + ydeltas[dir];

  if (!IN_MAP(obj->mapin, newx, newy))
    return (0);

  for (curr = obj; curr != NULL; curr = curr->link)
    if (MAP_MOVE_BLOCKED(obj->mapin, newx, newy, curr->z)
	&& !OB_IS_PHASED(obj))
      return (0);

  remove_ob(obj);

  if (OB_IS_PHASED(obj))
    insert_ob_under_map(obj->mapin, obj, newx, newy, obj->z);
  else
    insert_ob_in_map(obj->mapin, obj, newx, newy, obj->z);

  if (!IS_PLAYER(obj) && IS_ALIVE(obj))
    obj->facing = dir;
  return (1);
}

int
calc_reduction(object * target, int damage, int type)
{
  int             tmp;

  switch (type)
  {
  case (physical):
    tmp = target->resist.physical + 7;
    break;
  case (electric):
    tmp = target->resist.electric + 7;
    break;
  case (mental):
    tmp = target->resist.mental + 7;
    break;
  case (heat):
    tmp = target->resist.heat + 7;
    break;
  case (cold):
    tmp = target->resist.cold + 7;
    break;
  case (acid):
    tmp = target->resist.acid + 7;
    break;
  case (radiation):
    tmp = target->resist.radiation + 7;
    break;
  case (poison):
    tmp = target->resist.poison + 7;
    break;
  default:
    tmp = 7;
  }
  return (damage * damage_reductions[tmp]) / 100;
}

int
damage_object(object * target, object * hitter, int damage, int type)
{
  int             damage2,
                  damage3 = 0;

  if (target == NULL)
    return 0;

  target = HEAD(target);
  
  if (IS_ALIVE(target))        /* calculate how much absorbed by
                * armour */
  {
    damage2 = 6 * (damage / 10);  /* less bits truncated? */
    if (AN_OB_ITEM(target).torso != NULL)
    {
      damage_object(AN_OB_ITEM(target).torso, hitter, damage2, type);
      damage2 += calc_reduction(AN_OB_ITEM(target).torso, damage2, type);
    }
    damage3 += damage2;

    damage2 = (damage / 10);
    if (AN_OB_ITEM(target).head != NULL)
    {
      damage_object(AN_OB_ITEM(target).head, hitter, damage2, type);
      damage2 += calc_reduction(AN_OB_ITEM(target).head, damage2, type);
    }
    damage3 += damage2;

    damage2 = 3 * (damage / 10);
    if (AN_OB_ITEM(target).legs != NULL)
    {
      damage_object(AN_OB_ITEM(target).legs, hitter, damage2, type);
      damage2 += calc_reduction(AN_OB_ITEM(target).legs, damage2, type);
    }
    damage3 += damage2;

    /* get back stuff lost to truncation */
    damage3 += damage - ((damage / 10) * 10);

  } else
    damage3 = damage;
  damage3 += calc_reduction(target, damage3, type);
 
  if (type == mental)
    if (GET_MENTAL(target) > 0)
      SET_MENTAL(target, GET_MENTAL(target) - damage3);
    else
      SET_PHYS(target, GET_PHYS(target) - damage3);
  else
    SET_PHYS(target, GET_PHYS(target) - damage3);

  if (damage3 && IS_PLAYER(target))
    SET_PL_STATS_CHANGED(OB_PLDATA(target));

  SENDRQ(target, recieve_attack) (target, hitter, damage3,type);

  if (GET_PHYS(target) < 0 && !IS_PLAYER(target))
  {
    remove_ob(target);
    free_object(target);
  }
  return damage3;
}

int
attack_object(object * target, object * hitter, object * weapon)
{
  int             roll = random() % 100;

  if (target == NULL || distance(target, hitter) > 2)
    return (0);

  target = HEAD(target);
  
  if (IS_ALIVE(target))
    roll += OB_STATS(target).armour;
  if (GET_ACC(hitter) >= roll)
  {
    roll = damage_object(target, hitter,
            (rand() % OB_STATS(hitter).damage), physical);
    if (weapon != NULL)
      damage_object (weapon,NULL,roll,physical);
   }
  else
    roll = 0;
  
  return (roll);
}

int
check_use(object * user, object * obj)
{
  if (!IS_PLAYER(user) || PL_HAS_DM_POWER(OB_PLDATA(user)))
    return 1;
  switch (OB_PLDATA(user)->class)
  {
  case (mystic):
    return obj->usage.mystic;
    break;
  case (disciple):
    return obj->usage.disciple;
    break;
  case (ninja):
    return obj->usage.ninja;
    break;
  case (brute):
    return obj->usage.brute;
    break;
  case (warrior):
    return obj->usage.warrior;
    break;
  case (treach):
    return obj->usage.treach;
    break;
  case (sniper):
    return obj->usage.sniper;
    break;
  case (noclass):
    return obj->usage.none;
    break;
  default:
    fprintf (stderr,"error- player w/o class!");
    return 0;
    break;
  }
}

void
apply_bonuses(object * who, object * what)
{
  int             k;

 
  if (!IS_WEAPON(what)&& !IS_ARMOUR(what))
    return;
  for (k = 0; k < 4; k++)
  {
    switch (what->data.wecomp->bonuses[k].type)
    {
    case (nobonus):
      break;
    case (str):
      SET_STR(who, GET_STR(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (sma):
      SET_SMA(who, GET_SMA(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (wis):
      SET_WIS(who, GET_WIS(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (con):
      SET_CON(who, GET_CON(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (dex):
      SET_DEX(who, GET_DEX(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (luck):
      SET_LUK(who, GET_LUK(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (looks):
      SET_LOOKS(who, GET_LOOKS(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (cha):
      SET_CHA(who, GET_CHA(who) + what->data.wecomp->bonuses[k].amount);
      break;
    case (speed):
      SET_SPEED(who, GET_SPEED(who) + what->data.wecomp->bonuses[k].amount);
      break;
    default:
      break;

    }
  }
}

void
unapply_bonuses(object * who, object * what)
{
  int             k;

  if (!IS_WEAPON(what) && !IS_ARMOUR(what))
    return;
  for (k = 0; k < 4; k++)
  {
    switch (what->data.wecomp->bonuses[k].type)
    {
    case (nobonus):
      break;
    case (str):
      SET_STR(who, GET_STR(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (sma):
      SET_SMA(who, GET_SMA(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (wis):
      SET_WIS(who, GET_WIS(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (con):
      SET_CON(who, GET_CON(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (dex):
      SET_DEX(who, GET_DEX(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (luck):
      SET_LUK(who, GET_LUK(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (looks):
      SET_LOOKS(who, GET_LOOKS(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (cha):
      SET_CHA(who, GET_CHA(who) - what->data.wecomp->bonuses[k].amount);
      break;
    case (speed):
      SET_SPEED(who, GET_SPEED(who) - what->data.wecomp->bonuses[k].amount);
      break;
    default:
      break;
    }
  }

}


void
equip_weapon(object * who, object * what)
{
  SET_DAM(who, GET_DAM(who) + GET_WEAP_DAM(what));
  SET_ACC(who, GET_ACC(who) + GET_WEAP_ACC(what));
  apply_bonuses(who, what);
  AN_OB_ITEM(who).held = what;
}

void
unequip_weapon(object * what)
{
  if (what->owner != NULL && AN_OB_ITEM(what->owner).held == what)
  {
    SET_DAM(what->owner, GET_DAM(what->owner) -
      GET_WEAP_DAM(AN_OB_ITEM(what->owner).held));
    SET_ACC(what->owner, GET_ACC(what->owner) -
      GET_WEAP_ACC(AN_OB_ITEM(what->owner).held));
    unapply_bonuses(what->owner, what);
    AN_OB_ITEM(what->owner).held = NULL;
  }
}

void
equip_armour(object * who, object * what)
{
  SET_ARMOUR(who, GET_ARMOUR(who) + GET_ARMOR_VAL(what));
  SET_ACC(who, GET_ACC(who) + GET_ARM_ACC(what));
  apply_bonuses(who, what);
  switch (what->subtype)
  {
  case (torso):
    AN_OB_ITEM(what->owner).torso = what;
  case (head):
    AN_OB_ITEM(what->owner).head = what;
  case (legs):
    AN_OB_ITEM(what->owner).legs = what;
  case (neck):
    AN_OB_ITEM(what->owner).neck = what;
  case (finger):
    if (AN_OB_ITEM(what->owner).right_finger == NULL)
      AN_OB_ITEM(what->owner).right_finger = what;
    else
      AN_OB_ITEM(what->owner).left_finger = what;
    break;
  default:
    break;
  }
}

void
unequip_armour(object * what)
{
  if (what->owner == NULL)
    return;

  SET_ARMOUR(what->owner, GET_ARMOUR(what->owner) - GET_ARMOR_VAL(what));
  SET_ACC(what->owner, GET_ACC(what->owner) - GET_ARM_ACC(what));
  unapply_bonuses(what->owner, what);
  switch (what->subtype)
  {
  case (torso):
    AN_OB_ITEM(what->owner).torso = NULL;
  case (head):
    AN_OB_ITEM(what->owner).head = NULL;
  case (legs):
    AN_OB_ITEM(what->owner).legs = NULL;
  case (neck):
    AN_OB_ITEM(what->owner).neck = NULL;
  case (finger):
    if (AN_OB_ITEM(what->owner).right_finger == what)
      AN_OB_ITEM(what->owner).right_finger = NULL;
    else
      AN_OB_ITEM(what->owner).left_finger = NULL;
    break;
  default:
    break;
  }
}

void
unequip_held(object * who)
{
  if (AN_OB_ITEM(who).held != NULL)
  {
    CLEAR_IS_EQUIPPED(AN_OB_ITEM(who).held);
    switch (GET_TYPE(AN_OB_ITEM(who).held))
    {
      case (weapon):
      unequip_weapon(AN_OB_ITEM(who).held);
      break;
    default:
      break;
    }
    AN_OB_ITEM(who).held = NULL;
  }
}

#define NO_OWN 1
void
unequip_object(object * who, object * what)
{
  if (!IS_ALIVE(who))
  {
    fprintf(stderr, "non-alive:%s attempted equip!", GET_NAME(who));
    return;
  }
  if (what->owner != who)
  {
  /*  if (IS_PLAYER(who))
      write_text(OB_PLDATA(who), "You must own it to unequip it!");*/
    return;
  }
  switch (what->type)
  {
  case (player):
    return;
    break;
  case (animate):
    return;
    break;
  case (inanimate):
    break;
  case (weapon):
    unequip_weapon(what);
    break;
  case (armour):
    unequip_armour(what);
    break;
  default:
    return;
    break;
  }
  CLEAR_IS_EQUIPPED(what);
}

#define NO_USE 2
int
equip_object(object * who, object * what)
{
  if (what == NULL)
    return 0;
  if (!IS_ALIVE(who))
  {
    fprintf(stderr, "non-alive:%s attempted equip!", GET_NAME(who));
    return 0;
  }
  if (what->owner != who)
  {
  /*  if (IS_PLAYER(who))
      write_text(OB_PLDATA(who), "You must own it to equip it!");*/
    return NO_OWN;
  }
  if (!CAN_USE(who, what))
  {
  /* if (IS_PLAYER(who))
      write_text(OB_PLDATA(who), "You Cannot Use That...");*/
    return NO_USE;
  }
  switch (what->type)
  {
  case (player):
    return 0;
    break;
  case (animate):
    return 0;
    break;
  case (inanimate):
    if (AN_OB_ITEM(who).held == what)
    {
      unequip_held(who);
      return 0;
    } else
    {
      unequip_held(who);
      AN_OB_ITEM(who).held = what;
    }
  case (weapon):
    if (AN_OB_ITEM(who).held == what)
    {
      unequip_weapon(what);
      CLEAR_IS_EQUIPPED(what);
      return 0;
    } else
    {
      unequip_held(who);
      equip_weapon(who, what);
    }
    break;
  case (armour):
    switch (what->subtype)
    {
    case (torso):
      if (AN_OB_ITEM(who).torso == what)
        unequip_object(who, what);
      return 0;
      break;
    case (neck):
      if (AN_OB_ITEM(who).neck == what)
        unequip_object(what->owner, what);
      return 0;
      break;
    case (legs):
      if (AN_OB_ITEM(who).legs == what)
        unequip_object(who, what);
      return 0;
      break;
    case (head):
      if (AN_OB_ITEM(who).head == what)
        unequip_object(who, what);
      return 0;
      break;
    case (finger):
      if (AN_OB_ITEM(who).right_finger == what
          || AN_OB_ITEM(who).left_finger == what)
        unequip_object(who, what);
      return 0;
      break;
    default:
      return 0;
      break;
    }
    equip_armour(who, what);
    break;
  default:
    return 0;
    break;
  }
  SET_IS_EQUIPPED(what);
  return 1;
}

void
pickup_object(object * who, object * what)
{
  if (what == NULL || !CAN_TAKE(what))
    return;
  if (distance(who, what) > 2)
    return;
  if (what->owner != NULL)
    remove_held(what);
  else
    remove_ob(what);
  add_holding(who, what);
}

void
drop_object(object * who, object * what)
{
  object         *current;

  if (what == NULL || !CAN_DROP(what))
    return;

  /* find the ultimate owner */
  for (current = what; current->owner != NULL; current = current->owner);

  if (current != who || current == what)
    return;

  if (IS_EQUIPPED(what))
  {

    /*
     * if (IS_PLAYER(who)) write_text(OB_PLDATA(who),"You must
     * Unequip it first before dropping"); return;
     */
    unequip_object(who, what);
  }
  if (what->owner != who)
  {
    remove_held(what);
    add_holding(what->owner, what);
  } else
  {
    remove_held(what);
    insert_ob_under_map(who->mapin, what, who->x, who->y, who->z);
  }
}

void
set_object_dir(object *ob,int dir)
{
  object *tmp;
  
  for (tmp = ob; tmp != NULL; tmp = tmp->link)
    tmp->facing = dir;
}


void
send_first_channel (object *sender,int data1,void *data2)
{
  channel_entry *entry;
  
  if (sender->channels == NULL)
    return;

  for (entry = sender->channels->head->entries;
       entry != NULL; entry = entry->next)
    SENDRQ(entry->obj,channel)(entry->obj,sender,data1,data2);
}
    
void
send_all_channels (object *sender,int data1,void *data2)
{
  channel_ref *ref;
  channel_entry *entry;

 
  for (ref = sender->channels; ref != NULL; ref = ref->next)

    for (entry = ref->head->entries; entry != NULL; entry = entry->next)
      SENDRQ(entry->obj,channel)(entry->obj,sender,data1,data2);
    
}
